
class Biome:
    """
    Represents a minecraft biome.

    Attributes
    ----------
    namespace: :class:`str`
        Namespace of the biome, most of the time this is ``minecraft``
    id: :class:`str`
        ID of the biome, for example: forest, warm_ocean, etc...
    """
    __slots__ = ('namespace', 'id')

    def __init__(self, namespace: str, biome_id: str=None):
        """
        Parameters
        ----------
        namespace
            Namespace of the biome. If no biome_id is given, assume this is ``biome_id`` and set namespace to ``"minecraft"``
        biome_id
            ID of the biome
        """
        if biome_id is None:
            self.namespace = 'minecraft'
            self.id = namespace
        else:
            self.namespace = namespace
            self.id = biome_id

    def name(self) -> str:
        """
        Returns the biome in the ``minecraft:biome_id`` format
        """
        return self.namespace + ':' + self.id

    def __repr__(self):
        return f'Biome({self.name()})'

    def __eq__(self, other):
        if not isinstance(other, Biome):
            return False
        return self.namespace == other.namespace and self.id == other.id

    def __hash__(self):
        return hash(self.name())

    @classmethod
    def from_name(cls, name: str):
        """
        Creates a new Biome from the format: ``namespace:biome_id``

        Parameters
        ----------
        name
            Biome in said format
        """
        namespace, biome_id = name.split(':')
        return cls(namespace, biome_id)



"""
Create new section
"""
from struct import Struct
import array
from typing import List, Tuple, Union

from .block import Block
from .biome import Biome
from .errors import OutOfBoundsCoordinates
from . import nbt


# dirty mixin to change q to Q
def _update_fmt(self, length):
    self.fmt = Struct(f'>{length}Q')

nbt.TAG_Long_Array.update_fmt = _update_fmt


def bin_append(a, b, length=None):
    length = length or b.bit_length()
    return (a << length) | b


class EmptySection:
    """
    Used for making own sections.

    This is where the blocks are actually stored, in a 16³ sized array.
    To save up some space, ``None`` is used instead of the air block object,
    and will be replaced with ``self.air`` when needed

    Attributes
    ----------
    y: :class:`int`
        Section's Y index
    blocks: List[:class:`Block`]
        1D list of blocks
    air: :class:`Block`
        An air block
    """
    air = Block('minecraft', 'air')

    __slots__ = ('y', 'blocks', 'biomes')
    def __init__(self, y: int):
        self.y = y
        # 16 * 16 * 16 = 4096 blocks
        self.blocks: List[Union[Block, None]] = [None] * 4096
        # 4 * 4 * 4 = 64 biomes
        self.biomes: List[Union[Biome, None]] = [None] * 64

    @staticmethod
    def inside(x: int, y: int, z: int) -> bool:
        """
        Check if X Y and Z are in range of 0-15
        
        Parameters
        ----------
        int x, y, z
            Coordinates
        """
        return x >= 0 and x <= 15 and y >= 0 and y <= 15 and z >= 0 and z <= 15

    def set_block(self, block: Block, x: int, y: int, z: int):
        """
        Sets the block at given coordinates
        
        Parameters
        ----------
        block
            Block to set
        int x, y, z
            Coordinates

        Raises
        ------
        anvil.OutOfBoundsCoordinates
            If coordinates are not in range of 0-15
        """
        if not self.inside(x, y, z):
            raise OutOfBoundsCoordinates('X Y and Z must be in range of 0-15')
        index = y * 256 + z * 16 + x
        self.blocks[index] = block

    def get_block(self, x: int, y: int, z: int) -> Block:
        """
        Gets the block at given coordinates.
        
        Parameters
        ----------
        int x, y, z
            Coordinates

        Raises
        ------
        anvil.OutOfBoundsCoordinates
            If coordinates are not in range of 0-15
        """
        if not self.inside(x, y, z):
            raise OutOfBoundsCoordinates('X Y and Z must be in range of 0-15')
        index = y * 256 + z * 16 + x
        return self.blocks[index] or self.air

    def palette(self) -> Tuple[Block]:
        """
        Generates and returns a tuple of all the different blocks in the section
        The order can change as it uses sets, but should be fine when saving since
        it's only called once.
        """
        palette = set(self.blocks)
        if None in palette:
            palette.remove(None)
            palette.add(self.air)
        return tuple(palette)

    def blockstates(self, palette: Tuple[Block]=None) -> array.array:
        """
        Returns a list of each block's index in the palette.
        
        This is used in the BlockStates tag of the section.

        Parameters
        ----------
        palette
            Section's palette. If not given will generate one.
        """
        palette = palette or self.palette()
        bits = max((len(palette) - 1).bit_length(), 4)
        states = array.array('Q')
        current = 0
        current_len = 0
        for block in self.blocks:
            if block is None:
                index = palette.index(self.air)
            else:
                index = palette.index(block)
            # since 1.16, index will not span across arrays
            # If it's more than 64 bits then add to list and start over
            # with the remaining bits from last one
            if current_len + bits > 64:
                # leftover = 64 - current_len
                # states.append(bin_append(index & ((1 << leftover) - 1), current, length=current_len))
                # current = index >> leftover
                # current_len = bits - leftover
                states.append(current)
                current = index
                current_len = bits
            else:
                current = bin_append(index, current, length=current_len)
                current_len += bits
        states.append(current)
        return states

    def set_biome(self, biome: Biome, x: int, y: int, z: int):
        """Sets biome"""
        if x not in (0, 1, 2, 3):
            raise OutOfBoundsCoordinates(f'X ({x!r}) must be in range of 0 to 3')
        if z not in (0, 1, 2, 3):
            raise OutOfBoundsCoordinates(f'Z ({z!r}) must be in range of 0 to 3')
        if y not in (0, 1, 2, 3):
            raise OutOfBoundsCoordinates(f'Y ({z!r}) must be in range of 0 to 3')

        index = y * 16 + z * 4 + x
        self.biomes[index] = biome

    def biome_data(self, palette: Tuple[Biome]) -> array.array:
        """
        Returns a list of each biome's index in the palette.
        
        This is used in the 'biomes' tag of the section.

        ----------
        palette
            Section's biome palette.
        """
        if len(palette) < 2:
            return array.array('Q', [0] * 64)
        if len(palette) > 63:
            raise ValueError('Biome palette is too large')

        bits = (len(palette) - 1).bit_length()
        states = array.array('Q')
        current = 0
        current_len = 0
        for biome in self.biomes:
            if biome is None:
                index = palette.index(palette[0])
            else:
                index = palette.index(biome)
            if current_len + bits > 64:
                # TODO: do biome index span across 64bit numbers?
                # leftover = 64 - current_len
                # states.append(bin_append(index & ((1 << leftover) - 1), current, length=current_len))
                # current = index >> leftover
                # current_len = bits - leftover
                states.append(current)
                current = index
                current_len = bits
            else:
                current = bin_append(index, current, length=current_len)
                current_len += bits
        states.append(current)
        return states

    def save(self) -> nbt.TAG_Compound:
        """
        Saves the section to a TAG_Compound and is used inside the chunk tag
        This is missing the SkyLight tag, but minecraft still accepts it anyway
        """
        root = nbt.TAG_Compound()
        root.tags.append(nbt.TAG_Byte(name='Y', value=self.y))

        palette = self.palette()
        block_states = nbt.TAG_Compound(name='block_states')
        nbt_pal = nbt.TAG_List(name='palette', type=nbt.TAG_Compound)
        for block in palette:
            tag = nbt.TAG_Compound()
            tag.tags.append(nbt.TAG_String(name='Name', value=block.name()))
            if block.properties:
                properties = nbt.TAG_Compound()
                properties.name = 'Properties'
                for key, value in block.properties.items():
                    if isinstance(value, str):
                        properties.tags.append(nbt.TAG_String(name=key, value=value))
                    elif isinstance(value, bool):
                        # booleans are a string saved as either 'true' or 'false'
                        properties.tags.append(nbt.TAG_String(name=key, value=str(value).lower()))
                    elif isinstance(value, int):
                        # ints also seem to be saved as a string
                        properties.tags.append(nbt.TAG_String(name=key, value=str(value)))
                    else:
                        # assume its a nbt tag and just append it
                        properties.tags.append(value)
                tag.tags.append(properties)
            nbt_pal.tags.append(tag)
        # root.tags.append(nbt_pal)

        states = self.blockstates(palette=palette)
        # bstates = nbt.TAG_Long_Array(name='BlockStates')
        bstates = nbt.TAG_Long_Array(name='data')
        bstates.value = states.tolist()
        block_states.tags.append(nbt_pal)
        block_states.tags.append(bstates)
        root.tags.append(block_states)

        # biomes
        biomes = nbt.TAG_Compound(name='biomes')
        biome_palette = nbt.TAG_List(name='palette', type=nbt.TAG_String)
        biome_set = tuple(set(self.biomes))
        if len(biome_set) == 1 and None in biome_set:
            biome_set = (Biome('minecraft', 'the_void'),)
        for biome in biome_set:
            if biome:
                biome_palette.tags.append(nbt.TAG_String(name='Name', value=biome.name()))
        biomes.tags.append(biome_palette)
        if len(biome_set) > 1:
            biome_data = self.biome_data(palette=biome_set)
            bdata = nbt.TAG_Long_Array(name='data')
            bdata.value = biome_data.tolist()
            biomes.tags.append(bdata)

        root.tags.append(biomes)
        
        return root



from typing import Union, List, BinaryIO
from io import BytesIO
import zlib
import math
from .empty_chunk import EmptyChunk
from .chunk import Chunk
from .empty_section import EmptySection
from .block import Block
from .errors import OutOfBoundsCoordinates

def from_inclusive(a, b):
    """Returns a range from a to b, including both endpoints"""
    c = int(b > a)*2-1
    return range(a, b+c, c)

class EmptyRegion:
    """
    Used for making own regions
    
    Attributes
    ----------
    chunks: List[:class:`anvil.EmptyChunk`]
        List of chunks in this region
    x: :class:`int`
    z: :class:`int`
    """
    __slots__ = ('chunks', 'x', 'z')
    def __init__(self, x: int, z: int):
        # Create a 1d list for the 32x32 chunks
        self.chunks: List[EmptyChunk] = [None] * 1024
        self.x = x
        self.z = z

    def inside(self, x: int, y: int, z: int, chunk: bool=False) -> bool:
        """
        Returns if the given coordinates are inside this region
        
        Parameters
        ----------
        int x, y, z
            Coordinates
        chunk
            Whether coordinates are global or chunk coordinates
        """
        factor = 32 if chunk else 512
        rx = x // factor
        rz = z // factor
        return not (rx != self.x or rz != self.z or y < -64 or y > 319)

    def get_chunk(self, x: int, z: int) -> EmptyChunk:
        """
        Returns the chunk at given chunk coordinates
        
        Parameters
        ----------
        int x, z
            Chunk's coordinates

        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If the chunk (x, z) is not inside this region

        :rtype: :class:`anvil.EmptyChunk`
        """
        if not self.inside(x, 0, z, chunk=True):
            raise OutOfBoundsCoordinates(f'Chunk ({x}, {z}) is not inside this region')
        return self.chunks[z % 32 * 32 + x % 32]

    def add_chunk(self, chunk: EmptyChunk):
        """
        Adds given chunk to this region.
        Will overwrite if a chunk already exists in this location

        Parameters
        ----------
        chunk: :class:`EmptyChunk`
        
        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If the chunk (x, z) is not inside this region
        """
        if not self.inside(chunk.x, 0, chunk.z, chunk=True):
            raise OutOfBoundsCoordinates(f'Chunk ({chunk.x}, {chunk.z}) is not inside this region')
        self.chunks[chunk.z % 32 * 32 + chunk.x % 32] = chunk

    def add_section(self, section: EmptySection, x: int, z: int, replace: bool=True):
        """
        Adds section to chunk at (x, z).
        Same as ``EmptyChunk.add_section(section)``

        Parameters
        ----------
        section: :class:`EmptySection`
            Section to add
        int x, z
            Chunk's coordinate
        replace
            Whether to replace section if it already exists in the chunk
        
        Raises
        ------
        anvil.OutOfBoundsCoordinates
            If the chunk (x, z) is not inside this region
        """
        if not self.inside(x, 0, z, chunk=True):
            raise OutOfBoundsCoordinates(f'Chunk ({x}, {z}) is not inside this region')
        chunk = self.chunks[z % 32 * 32 + x % 32]
        if chunk is None:
            chunk = EmptyChunk(x, z)
            self.add_chunk(chunk)
        chunk.add_section(section, replace)

    def set_block(self, block: Block, x: int, y: int, z: int):
        """
        Sets block at given coordinates.
        New chunk is made if it doesn't exist.

        Parameters
        ----------
        block: :class:`Block`
            Block to place
        int x, y, z
            Coordinates

        Raises
        ------
        anvil.OutOfBoundsCoordinates
            If the block (x, y, z) is not inside this region
        """
        if not self.inside(x, y, z):
            raise OutOfBoundsCoordinates(f'Block ({x}, {y}, {z}) is not inside this region')
        cx = x // 16
        cz = z // 16
        chunk = self.get_chunk(cx, cz)
        if chunk is None:
            chunk = EmptyChunk(cx, cz)
            self.add_chunk(chunk)
        chunk.set_block(block, x % 16, y, z % 16)

    def set_if_inside(self, block: Block, x: int, y: int, z: int):
        """
        Helper function that only sets
        the block if ``self.inside(x, y, z)`` is true
        
        Parameters
        ----------
        block: :class:`Block`
            Block to place
        int x, y, z
            Coordinates
        """
        if self.inside(x, y, z):
            self.set_block(block, x, y, z)

    def fill(self, block: Block, x1: int, y1: int, z1: int, x2: int, y2: int, z2: int, ignore_outside: bool=False):
        """
        Fills in blocks from
        ``(x1, y1, z1)`` to ``(x2, y2, z2)``
        in a rectangle.

        Parameters
        ----------
        block: :class:`Block`
        int x1, y1, z1
            Coordinates
        int x2, y2, z2
            Coordinates
        ignore_outside
            Whether to ignore if coordinates are outside the region

        Raises
        ------
        anvil.OutOfBoundsCoordinates
            If any of the coordinates are outside the region
        """
        if not ignore_outside:
            if not self.inside(x1, y1, z1):
                raise OutOfBoundsCoordinates(f'First coords ({x1}, {y1}, {z1}) is not inside this region')
            if not self.inside(x2, y2, z2):
                raise OutOfBoundsCoordinates(f'Second coords ({x2}, {y2}, {z2}) is not inside this region')

        for y in from_inclusive(y1, y2):
            for z in from_inclusive(z1, z2):
                for x in from_inclusive(x1, x2):
                    if ignore_outside:
                        self.set_if_inside(block, x, y, z)
                    else:
                        self.set_block(block, x, y, z)

    def save(self, file: Union[str, BinaryIO, None]=None) -> bytes:
        """
        Returns the region as bytes with
        the anvil file format structure,
        aka the final ``.mca`` file.

        Parameters
        ----------
        file
            Either a path or a file object, if given region
            will be saved there.
        """
        # Store all the chunks data as zlib compressed nbt data
        chunks_data = []
        for chunk in self.chunks:
            if chunk is None:
                chunks_data.append(None)
                continue
            chunk_data = BytesIO()
            if isinstance(chunk, Chunk):
                # nbt_data = nbt.NBTFile()
                # nbt_data.tags.append(nbt.TAG_Int(name='DataVersion', value=chunk.version))
                # nbt_data.tags.append(chunk.data)
                nbt_data = chunk.data
            else:
                nbt_data = chunk.save()
            nbt_data.write_file(buffer=chunk_data)
            chunk_data.seek(0)
            chunk_data = zlib.compress(chunk_data.read())
            chunks_data.append(chunk_data)

        # This is what is added after the location and timestamp header
        chunks_bytes = bytes()
        offsets = []
        for chunk in chunks_data:
            if chunk is None:
                offsets.append(None)
                continue
            # 4 bytes are for length, b'\x02' is the compression type which is 2 since its using zlib
            to_add = (len(chunk)+1).to_bytes(4, 'big') + b'\x02' + chunk

            # offset in 4KiB sectors
            sector_offset = len(chunks_bytes) // 4096
            sector_count = math.ceil(len(to_add) / 4096)
            offsets.append((sector_offset, sector_count))

            # Padding to be a multiple of 4KiB long
            to_add += bytes(4096 - (len(to_add) % 4096))
            chunks_bytes += to_add

        locations_header = bytes()
        for offset in offsets:
            # None means the chunk is not an actual chunk in the region
            # and will be 4 null bytes, which represents non-generated chunks to minecraft
            if offset is None:
                locations_header += bytes(4)
            else:
                # offset is (sector offset, sector count)
                locations_header += (offset[0] + 2).to_bytes(3, 'big') + offset[1].to_bytes(1, 'big')

        # Set them all as 0
        timestamps_header = bytes(4096)

        final = locations_header + timestamps_header + chunks_bytes

        # Pad file to be a multiple of 4KiB in size
        # as Minecraft only accepts region files that are like that
        final += bytes(4096 - (len(final) % 4096))
        assert len(final) % 4096 == 0 # just in case

        # Save to a file if it was given
        if file:
            if isinstance(file, str):
                with open(file, 'wb') as f:
                    f.write(final)
            else:
                file.write(final)
        return final



import math
import time
import logging
import array
import pytest
from mca import EmptyRegion, Block, RawSection

LOGGER = logging.getLogger(__name__)

def test_benchmark():
    region = EmptyRegion(0, 0)

    block = Block('iron_block')

    def func(x: int, z: int) -> int:
        return math.sin(x * x + z * z) * 0.5 + 0.5

    w = 256
    scale = 0.05
    y_scale = 15

    start = time.time()
    for x in range(w):
        for z in range(w):
            u = x - w / 2
            v = z - w / 2
            y = int(func(u * scale, v * scale) * y_scale)
            region.set_block(block, x, y, z)
    end = time.time()

    LOGGER.info(f'Generating took: {end - start:.3f}s')

    times = []
    n = 3
    for _ in range(n):
        start = time.time()
        region.save()
        end = time.time()
        times.append(end - start)

    LOGGER.info(f'Saving (average of {n}) took: {sum(times) / len(times):.3f}s')

@pytest.mark.skip(reason='RawSection is deprecated')
def test_raw_section():
    region = EmptyRegion(0, 0)

    block = Block('iron_block')
    air = Block('air')
    palette = (air, block)

    def func(x: int, z: int) -> int:
        return math.sin(x * x + z * z) * 0.5 + 0.5

    w = 256
    chunk_w = w // 16
    scale = 0.05
    y_scale = 15

    start = time.time()

    # from 0 to y_scale
    heights = array.array('B')
    for z in range(w):
        for x in range(w):
            u = z - w / 2
            v = x - w / 2
            height = int(func(u * scale, v * scale) * y_scale)
            heights.append(height)
            
    for chunk_x in range(chunk_w):
        for chunk_z in range(chunk_w):
            blocks = array.array('B')
            for y in range(16):
                for z in range(16):
                    for x in range(16):
                        rx = x + chunk_x * 16
                        rz = z + chunk_z * 16
                        i = rz * w + rx
                        height = heights[i]
                        if y == height:
                            blocks.append(1)
                        else:
                            blocks.append(0)
            region.add_section(RawSection(0, blocks, palette), chunk_x, chunk_z)

    end = time.time()

    LOGGER.info(f'Generating took: {end - start:.3f}s')

    times = []
    n = 3
    for _ in range(n):
        start = time.time()
        region.save()
        end = time.time()
        times.append(end - start)

    LOGGER.info(f'Saving (average of {n}) took: {sum(times) / len(times):.3f}s')


@pytest.mark.skip(reason='RawSection is deprecated')
def test_raw_section_simple():
    region = EmptyRegion(0, 0)

    air = Block('air')
    palette = (air, Block('white_concrete'), Block('light_gray_concrete'), Block('gray_concrete'))

    blocks = array.array('B')
    for z in range(16):
        for x in range(16):
            d = x * x + z * z
            if d < 25:
                blocks.append(1)
            elif d < 100:
                blocks.append(2)
            elif d < 225:
                blocks.append(3)
            else:
                blocks.append(0)

    blocks.extend(0 for _ in range(16 * 16 * 16 - len(blocks)))

    assert len(blocks) == 16 * 16 * 16
    
    section = RawSection(0, blocks, palette)
    region.add_section(section, 0, 0)

    region.save()



from mca import Region
import io
import secrets

def test_from_filename(tmp_path):
    filename = tmp_path / "region.mca"
    contents = secrets.token_bytes()

    with open(filename, 'wb') as f:
        f.write(contents)

    region = Region.from_file(str(filename))
    assert region.data == contents

def test_from_filelike():
    contents = secrets.token_bytes()
    filelike = io.BytesIO(contents)

    region = Region.from_file(filelike)
    assert region.data == contents



class OutOfBoundsCoordinates(ValueError):
    """Exception used for when coordinates are out of bounds"""

class ChunkNotFound(Exception):
    """Exception used for when a chunk was not found"""

class EmptySectionAlreadyExists(Exception):
    """
    Exception used for when trying to add an `EmptySection` to an `EmptyChunk`
    and the chunk already has a section with the same Y
    """

class GZipChunkData(Exception):
    """Exception used when trying to get chunk data compressed in gzip"""

class DataNotAvailable(Exception):
    """Exception used when the data is not available"""


import json
from . import nbt

class Block:
    """
    Represents a minecraft block.

    Attributes
    ----------
    namespace: :class:`str`
        Namespace of the block, most of the time this is ``minecraft``
    id: :class:`str`
        ID of the block, for example: stone, diamond_block, etc...
    properties: :class:`dict`
        Block properties as a dict
    """
    __slots__ = ('namespace', 'id', 'properties')

    def __init__(self, namespace: str, block_id: str=None, properties: dict=None):
        """
        Parameters
        ----------
        namespace
            Namespace of the block. If no block_id is given, assume this is ``block_id`` and set namespace to ``"minecraft"``
        block_id
            ID of the block
        properties
            Block properties
        """
        if block_id is None:
            self.namespace = 'minecraft'
            self.id = namespace
        else:
            self.namespace = namespace
            self.id = block_id
        self.properties = properties or {}

    def name(self) -> str:
        """
        Returns the block in the ``minecraft:block_id`` format
        """
        return self.namespace + ':' + self.id

    def __repr__(self):
        return f'Block({self.name()})'

    def __eq__(self, other):
        if not isinstance(other, Block):
            return False
        return self.namespace == other.namespace and self.id == other.id and self.properties == other.properties

    def __hash__(self):
        return hash(self.name()) ^ hash(json.dumps(self.properties, sort_keys=True))

    @classmethod
    def from_name(cls, name: str, *args, **kwargs):
        """
        Creates a new Block from the format: ``namespace:block_id``

        Parameters
        ----------
        name
            Block in said format
        , args, kwargs
            Will be passed on to the main constructor
        """
        if ':' in name:
            namespace, block_id = name.split(':')
        else:
            namespace = 'minecraft'
            block_id = name
        return cls(namespace, block_id, *args, **kwargs)

    @classmethod
    def from_palette(cls, tag: nbt.TAG_Compound):
        """
        Creates a new Block from the tag format on Section.Palette

        Parameters
        ----------
        tag
            Raw tag from a section's palette
        """
        name = tag['Name'].value
        properties = tag.get('Properties')
        if properties:
            properties = dict(properties)
        return cls.from_name(name, properties=properties)



"""
TODO: will remove RawSection
"""
from typing import Sequence, Iterable
from . import EmptySection, Block
import array
from . import nbt

def bin_append(a, b, length=None):
    length = length or b.bit_length()
    return (a << length) | b

class RawSection:
    """
    Same as :class:`EmptySection` but you manually
    set the palette and the blocks array (which instead
    of :class:`Block`, it's indexes on the palette)
    
    Attributes
    ----------
    y: :class:`int`
        Section's Y index
    blocks: Iterable[:class:`int`]
        Array of palette indexes
    _palette: Sequence[:class:`Block`]
        Section's palette
    """
    __slots__ = ('y', '_palette', 'blocks')
    def __init__(self, y: int, blocks: Iterable[int], palette: Sequence[Block]):
        self.y = y
        self.blocks: Iterable[int] = blocks
        self._palette: Sequence[Block] = palette

    def palette(self) -> Sequence[Block]:
        """Returns ``self._palette``"""
        return self._palette

    def blockstates(self, palette: Sequence[Block]=None) -> array.array:
        """Refer to :class:`EmptySection.blockstates()`"""
        bits = max((len(self._palette) - 1).bit_length(), 4)
        states = array.array('Q')
        current = 0
        current_len = 0
        for index in self.blocks:
            if current_len + bits > 64:
                leftover = 64 - current_len
                states.append(bin_append(index & ((1 << leftover) - 1), current, length=current_len))
                current = index >> leftover
                current_len = bits - leftover
            else:
                current = bin_append(index, current, length=current_len)
                current_len += bits
        states.append(current)
        return states

    def save(self) -> nbt.TAG_Compound:
        """Refer to :class:`EmptySection.save()`"""
        return EmptySection.save(self)



from typing import Tuple, Union, BinaryIO
import zlib
from io import BytesIO
from . import nbt
from .chunk import Chunk
from .errors import GZipChunkData, ChunkNotFound

class Region:
    """
    Read-only region

    Attributes
    ----------
    data: :class:`bytes`
        Region file (``.mca``) as bytes
    """
    __slots__ = ('data',)
    def __init__(self, data: bytes):
        """Makes a Region object from data, which is the region file content"""
        self.data = data

    @staticmethod
    def header_offset(chunk_x: int, chunk_z: int) -> int:
        """
        Returns the byte offset for given chunk in the header
        
        Parameters
        ----------
        chunk_x
            Chunk's X value
        chunk_z
            Chunk's Z value
        """
        return 4 * (chunk_x % 32 + chunk_z % 32 * 32)

    def chunk_location(self, chunk_x: int, chunk_z: int) -> Tuple[int, int]:
        """
        Returns the chunk offset in the 4KiB sectors from the start of the file,
        and the length of the chunk in sectors of 4KiB

        Will return ``(0, 0)`` if chunk hasn't been generated yet

        Parameters
        ----------
        chunk_x
            Chunk's X value
        chunk_z
            Chunk's Z value
        """
        b_off = self.header_offset(chunk_x, chunk_z)
        off = int.from_bytes(self.data[b_off : b_off + 3], byteorder='big')
        sectors = self.data[b_off + 3]
        return (off, sectors)

    def chunk_data(self, chunk_x: int, chunk_z: int) -> nbt.NBTFile:
        """
        Returns the NBT data for a chunk
        
        Parameters
        ----------
        chunk_x
            Chunk's X value
        chunk_z
            Chunk's Z value

        Raises
        ------
        anvil.GZipChunkData
            If the chunk's compression is gzip
        """
        off = self.chunk_location(chunk_x, chunk_z)
        # (0, 0) means it hasn't generated yet, aka it doesn't exist yet
        if off == (0, 0):
            return
        off = off[0] * 4096
        length = int.from_bytes(self.data[off:off + 4], byteorder='big')
        compression = self.data[off + 4] # 2 most of the time
        if compression == 1:
            raise GZipChunkData('GZip is not supported')
        compressed_data = self.data[off + 5 : off + 5 + length - 1]
        return nbt.NBTFile(buffer=BytesIO(zlib.decompress(compressed_data)))

    def get_chunk(self, chunk_x: int, chunk_z: int) -> Chunk:
        """
        Returns the chunk at given coordinates,
        same as doing ``Chunk.from_region(region, chunk_x, chunk_z)``

        Parameters
        ----------
        chunk_x
            Chunk's X value
        chunk_z
            Chunk's Z value
        
        
        :rtype: :class:`anvil.Chunk`
        """
        nbt_data = self.chunk_data(chunk_x, chunk_z)
        if nbt_data is None:
            raise ChunkNotFound(f'Could not find chunk ({chunk_x}, {chunk_z})')
        return Chunk(nbt_data)

    @classmethod
    def from_file(cls, file: Union[str, BinaryIO]):
        """
        Creates a new region with the data from reading the given file

        Parameters
        ----------
        file
            Either a file path or a file object
        """
        if isinstance(file, str):
            with open(file, 'rb') as f:
                return cls(data=f.read())
        else:
            return cls(data=file.read())



from pytest import mark
from mca import Chunk, EmptyChunk, EmptyRegion, Block, Region
import random

def test_mixed_chunk_types():
    colors = ['stone', 'dirt', 'grass', 'water']

    region = EmptyRegion(0, 0)

    chunk = EmptyChunk(0, 0)
    empty_chunk = EmptyChunk(1, 0)
    for i, color in enumerate(colors):
        chunk.set_block(Block(color), i, 0, 0)
        empty_chunk.set_block(Block(color), i, 0, 0)

    region.add_chunk(Chunk(chunk.save()))
    region.add_chunk(empty_chunk)
    #region.add_chunk(Chunk(empty_chunk.save()))

    region = Region(region.save())

    for i in range(2):
        chunk = region.get_chunk(i, 0)
        for i, color in enumerate(colors):
            assert chunk.get_block(i, 0, 0).id == color



"""
Create empty chunk
"""
from typing import List, Union
from .biome import Biome
from .block import Block
from .empty_section import EmptySection
from .errors import OutOfBoundsCoordinates, EmptySectionAlreadyExists
from . import nbt

WORLD_VERSION = 2844 # 21w43a Minecraft 1.18 snapshot 7
NUM_SECTIONS_PER_CHUNK = 24

class EmptyChunk:
    """
    Used for making own chunks

    Attributes
    ----------
    x: :class:`int`
        Chunk's X position
    z: :class:`int`
        Chunk's Z position
    sections: List[:class:`anvil.EmptySection`]
        List of all the sections in this chunk
    version: :class:`int`
        Chunk's DataVersion
    """

    __slots__ = ('x', 'z', 'sections', 'version')
    def __init__(self, x: int, z: int, version: int = WORLD_VERSION):
        self.x = x
        self.z = z
        self.version = version
        self.sections: List[Union[EmptySection,None]] = [None] * NUM_SECTIONS_PER_CHUNK

    def add_section(self, section: EmptySection, replace: bool = True):
        """
        Adds a section to the chunk

        Parameters
        ----------
        section
            Section to add
        replace
            Whether to replace section if one at same Y already exists
        
        Raises
        ------
        anvil.EmptySectionAlreadyExists
            If ``replace`` is ``False`` and section with same Y already exists in this chunk
        """
        if section.y < -4 or section.y + 4 > NUM_SECTIONS_PER_CHUNK:
            raise OutOfBoundsCoordinates('section Y is too high.')
        if self.sections[section.y + 4] and not replace:
            raise EmptySectionAlreadyExists(f'EmptySection (Y={section.y}) already exists in this chunk')
        self.sections[section.y + 4] = section

    def get_block(self, x: int, y: int, z: int) -> Block:
        """
        Gets the block at given coordinates
        
        Parameters
        ----------
        int x, z
            In range of 0 to 15
        y
            In range of -64 to 319

        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If X, Y or Z are not in the proper range

        Returns
        -------
        block : :class:`anvil.Block` or None
            Returns ``None`` if the section is empty, meaning the block
            is most likely an air block.
        """
        if x < 0 or x > 15:
            raise OutOfBoundsCoordinates(f'X ({x!r}) must be in range of 0 to 15')
        if z < 0 or z > 15:
            raise OutOfBoundsCoordinates(f'Z ({z!r}) must be in range of 0 to 15')
        if y < -64 or y > 319:
            raise OutOfBoundsCoordinates(f'Y ({y!r}) must be in range of 0 to 255')
        section = self.sections[ (y + 64) // 16]
        if section is None:
            return None
        return section.get_block(x, y % 16, z)

    def set_block(self, block: Block, x: int, y: int, z: int):
        """
        Sets block at given coordinates
        
        Parameters
        ----------
        int x, z
            In range of 0 to 15
        y
            In range of -64 to 319

        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If X, Y or Z are not in the proper range
        
        """
        if x < 0 or x > 15:
            raise OutOfBoundsCoordinates(f'X ({x!r}) must be in range of 0 to 15')
        if z < 0 or z > 15:
            raise OutOfBoundsCoordinates(f'Z ({z!r}) must be in range of 0 to 15')
        if y < -64 or y > 319:
            raise OutOfBoundsCoordinates(f'Y ({y!r}) must be in range of 0 to 255')
        section = self.sections[ (y + 64) // 16]
        if section is None:
            section = EmptySection( y // 16)
            self.add_section(section)
        section.set_block(block, x, y % 16, z)


    def set_biome(self, biome: Biome, x: int, y: int, z: int):
        """
        Sets biome at given coordinates
        
        Parameters
        ----------
        int x, z
            In range of 0 to 15
        y
            In range of -64 to 319

        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If X, Y or Z are not in the proper range
        
        """
        if x < 0 or x > 15:
            raise OutOfBoundsCoordinates(f'X ({x!r}) must be in range of 0 to 15')
        if z < 0 or z > 15:
            raise OutOfBoundsCoordinates(f'Z ({z!r}) must be in range of 0 to 15')
        if y < -64 or y > 319:
            raise OutOfBoundsCoordinates(f'Y ({y!r}) must be in range of 0 to 255')
        section = self.sections[ (y + 64) // 16]
        if section is None:
            section = EmptySection( y // 16)
            self.add_section(section)
        section.set_biome(biome, x // 4, (y % 16) // 4, z // 4)

    def save(self) -> nbt.NBTFile:
        """
        Saves the chunk data to a :class:`NBTFile`

        Notes
        -----
        Does not contain most data a regular chunk would have,
        but minecraft stills accept it.
        """
        root = nbt.NBTFile()
        root.tags.append(nbt.TAG_Int(name='DataVersion',value=self.version))
        #level = nbt.TAG_Compound()
        # Needs to be in a separate line because it just gets
        # ignored if you pass it as a kwarg in the constructor
        root.tags.extend([
            nbt.TAG_List(name='Entities', type=nbt.TAG_Compound),
            nbt.TAG_List(name='block_entities', type=nbt.TAG_Compound),
            nbt.TAG_List(name='LiquidTicks', type=nbt.TAG_Compound),
            nbt.TAG_Int(name='xPos', value=self.x),
            nbt.TAG_Int(name='zPos', value=self.z),
            nbt.TAG_Long(name='LastUpdate', value=0),
            nbt.TAG_Long(name='InhabitedTime', value=0),
            nbt.TAG_Byte(name='isLightOn', value=1),
            nbt.TAG_String(name='Status', value='full')
        ])
        sections = nbt.TAG_List(name='sections', type=nbt.TAG_Compound)
        for s in self.sections:
            if s:
                p = s.palette()
                # Minecraft does not save sections that are just air
                # So we can just skip them
                if len(p) == 1 and p[0].name() == 'minecraft:air':
                    continue
                sections.tags.append(s.save())
        root.tags.append(sections)
        return root



from pytest import mark
from mca import Biome, Chunk, EmptyChunk, EmptyRegion, Block, Region


def test_basic_biome():
    blocks = ['stone', 'dirt', 'grass', 'water']
    biomes = ['plains', 'desert', 'mountains', 'ocean']

    region = EmptyRegion(0, 0)

    chunk = EmptyChunk(0, 0)
    for i in range(4):
        chunk.set_block(Block(blocks[i]), i, 0, 0)
        chunk.set_biome(Biome(biomes[i]), 0, i * 4, 0)

    region.add_chunk(Chunk(chunk.save()))
    #region.add_chunk(Chunk(empty_chunk.save()))

    region = Region(region.save())

    #for i in range(2):
    #    chunk = region.get_chunk(i, 0)
    #    for i, color in enumerate(colors):
    #        assert chunk.get_block(i, 0, 0).id == color
    chunk2 = region.get_chunk(0, 0)
    for i in range(4):
        assert chunk2.get_block(i, 0, 0).id == blocks[i]
        # TODO: get_biome doesn't work yet
        # assert chunk2.get_biome(0, i * 4, 0).id == biomes[i]



import setuptools

with open('README.md', 'r') as file:
    long_description = file.read()

setuptools.setup(
    name='mcapy',
    version='0.9.1',
    author='wensheng',
    description='A Minecraft MCA file reader/writer',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/wensheng/mcapy',
    packages=['mca'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[],
    include_package_data=True
)



from .chunk import Chunk
from .biome import Biome
from .block import Block
from .region import Region
from .empty_region import EmptyRegion
from .empty_chunk import EmptyChunk
from .empty_section import EmptySection
from .raw_section import RawSection
from . import nbt



from pytest import mark
from mca import EmptyRegion, Region, Block

def coord_to_index(x, y, z):
    return y * 16 * 16 + z * 16 + x

def test_4bits():
    region = EmptyRegion(0, 0)

    region.set_block(Block('minecraft', 'stone'), 0, 0, 0)
    region.set_block(Block('minecraft', 'dirt'), 1, 0, 0)
    region.set_block(Block('minecraft', 'oak_planks'), 2, 0, 0)
    region.set_block(Block('minecraft', 'sand'), 10, 7, 5)
    region.set_block(Block('minecraft', 'white_wool'), 8, 6, 0)
    region.set_block(Block('minecraft', 'bedrock'), 15, 15, 15)

    region = Region(region.save())

    for i, block in enumerate(region.get_chunk(0, 0).stream_blocks()):
        if i == 0:
            assert block.id == 'stone'
        elif i == 1:
            assert block.id == 'dirt'
        elif i == 2:
            assert block.id == 'oak_planks'
        elif i == coord_to_index(10, 7, 5):
            assert block.id == 'sand'
        elif i == coord_to_index(15, 15, 15):
            assert block.id == 'bedrock'
        elif i == coord_to_index(8, 6, 0):
            assert block.id == 'white_wool'
        else:
            assert block.id == 'air'

# @mark.skip(reason="temp")
def test_5bits():
    from random import randint
    region = EmptyRegion(0, 0)

    blocks = """stone dirt oak_planks sand bedrock white_wool red_wool
                green_wool sponge awesome these dont need to exist
                foo bar""".split()
    positions = list(set(
        [(randint(0, 15), randint(0, 15), randint(0, 15)) for _ in range(len(blocks))]
    ))
    for i, pos in enumerate(positions):
        region.set_block(Block(blocks[i]), *pos)

    region = Region(region.save())

    for i, block in enumerate(region.get_chunk(0, 0).stream_blocks()):
        if block.id in blocks:
            assert coord_to_index(*positions[blocks.index(block.id)]) == i
        else:
            assert block.id == 'air'

# @mark.skip(reason="temp")
def test_index():
    region = EmptyRegion(0, 0)

    region.set_block(Block('minecraft', 'dirt'), 2, 0, 0)
    region.set_block(Block('minecraft', 'stone'), 3, 0, 0)
    blocks = 'stone,dirt,oak_planks,sand,bedrock'.split(',')
    for i, block in enumerate(blocks):
        region.set_block(Block('minecraft', block), i % 16, 0, i // 16)

    region = Region(region.save())

    for i, block in enumerate(region.get_chunk(0, 0).stream_blocks(index=2)):
        i += 2
        if i < len(blocks):
            assert block.id == blocks[i]
        else:
            assert block.id == 'air'

# @mark.skip(reason="temp")
def test_index_5bits():
    region = EmptyRegion(0, 0)

    region.set_block(Block('minecraft', 'dirt'), 2, 0, 0)
    region.set_block(Block('minecraft', 'stone'), 3, 0, 0)
    blocks = """stone dirt oak_planks sand bedrock white_wool red_wool green_wool
                sponge awesome these dont need to exist foo bar""".split()
    for i, block in enumerate(blocks):
        region.set_block(Block('minecraft', block), i % 16, 0, i // 16)

    region = Region(region.save())
    
    for i, block in enumerate(region.get_chunk(0, 0).stream_blocks(index=17)):
        i += 17
        if i < len(blocks):
            assert block.id == blocks[i]
        else:
            assert block.id == 'air'



"""
Chunk class
ref: https://minecraft.wiki/w/Chunk_format
"""
import sys
from typing import Union, Tuple, Generator, Optional

from . import nbt
from .block import Block
from .biome import Biome
from .errors import OutOfBoundsCoordinates, DataNotAvailable


def bin_append(a, b, length=None):
    """
    Appends number a to the left of b
    bin_append(0b1, 0b10) = 0b110
    """
    length = length or b.bit_length()
    return (a << length) | b


def nibble(byte_array, index):
    value = byte_array[index // 2]
    if index % 2:
        return value >> 4
    else:
        return value & 0b1111


class Chunk:
    """
    Represents a chunk from a ``.mca`` file.

    Note that this is read only.

    Attributes
    ----------
    x: :class:`int`
        Chunk's X position
    z: :class:`int`
        Chunk's Z position
    version: :class:`int`
        Version of the chunk NBT structure
    data: :class:`nbt.TAG_Compound`
        Raw NBT data of the chunk
    tile_entities: :class:`nbt.TAG_Compound`
        ``self.data['TileEntities']`` as an attribute for easier use
    """
    __slots__ = ('version', 'data', 'x', 'z', 'tile_entities')

    def __init__(self, nbt_data: nbt.NBTFile):
        try:
            self.version = nbt_data['DataVersion'].value
        except KeyError:
            sys.exit("mcapy does not work with Data Version<2844 (21w43a, or Minecraft 1.18 s7)")

        self.data = nbt_data
        self.x = self.data['xPos'].value
        self.z = self.data['zPos'].value
        self.tile_entities = self.data['block_entities']

    def get_section(self, y: int) -> nbt.TAG_Compound:
        """
        Returns the section at given y index
        can also return nothing if section is missing, aka it's empty

        Parameters
        ----------
        y
            Section Y index

        Raises
        ------
        anvil.OutOfBoundsCoordinates
            If Y is not in range of 0 to 15
        """
        if y < -4 or y > 19:
            raise OutOfBoundsCoordinates(f'Y ({y!r}) must be in range of -4 to 19')

        try:
            sections = self.data["sections"]
        except KeyError:
            return None

        for section in sections:
            if section['Y'].value == y:
                return section

    def get_palette(self, section: Union[int, nbt.TAG_Compound]) -> Tuple[Block]:
        """
        Returns the block palette for given section

        Parameters
        ----------
        section
            Either a section NBT tag or an index


        :rtype: Tuple[:class:`anvil.Block`]
        """
        if isinstance(section, int):
            section = self.get_section(section)
        if section is None:
            return
        return tuple(Block.from_palette(i) for i in section['Palette'])

    def get_block(self, x: int, y: int, z: int, section: Union[int, nbt.TAG_Compound]=None) -> Block:
        """
        Returns the block in the given coordinates

        Parameters
        ----------
        int x, y, z
            Block's coordinates in the chunk
        section : int
            Either a section NBT tag or an index. If no section is given,
            assume Y is global and use it for getting the section.

        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If X, Y or Z are not in the proper range

        :rtype: :class:`anvil.Block`
        """
        if x < 0 or x > 15:
            raise OutOfBoundsCoordinates(f'X ({x!r}) must be in range of 0 to 15')
        if z < 0 or z > 15:
            raise OutOfBoundsCoordinates(f'Z ({z!r}) must be in range of 0 to 15')
        if y < -64 or y > 319:
            raise OutOfBoundsCoordinates(f'Y ({y!r}) must be in range of -64 to 319')

        if section is None:
            section = self.get_section(y // 16)
            # global Y to section Y
            y %= 16

        # If its an empty section its most likely an air block
        if section is None or 'block_states' not in section:
            return Block.from_name('minecraft:air')

        # Number of bits each block is on BlockStates
        # Cannot be lower than 4
        bits = max((len(section['block_states']['palette']) - 1).bit_length(), 4)

        # Get index on the block list with the order YZX
        index = y * 16 * 16 + z * 16 + x

        # BlockStates is an array of 64 bit numbers
        # that holds the blocks index on the palette list
        states = section['block_states']['data']

        state = index // (64 // bits)

        # makes sure the number is unsigned
        # by adding 2^64
        # could also use ctypes.c_ulonglong(n).value but that'd require an extra import
        data = states[state]
        if data < 0:
            data += 2**64

        shifted_data = data >> (index % (64 // bits) * bits)

        # get `bits` least significant bits
        # which are the palette index
        palette_id = shifted_data & 2**bits - 1

        block = section['block_states']['palette'][palette_id]
        return Block.from_palette(block)

    def stream_blocks(self, index: int=0, section: Union[int, nbt.TAG_Compound]=None) -> Generator[Block, None, None]:
        """
        Returns a generator for all the blocks in given section

        Parameters
        ----------
        index
            At what block to start from.

            To get an index from (x, y, z), simply do:

            ``y * 256 + z * 16 + x``
        section
            Either a Y index or a section NBT tag.

        Raises
        ------
        anvil.OutOfBoundCoordidnates
            If `section` is not in the range of 0 to 15

        Yields
        ------
        :class:`anvil.Block`
        """
        if isinstance(section, int) and (section < -4 or section > 19):
            raise OutOfBoundsCoordinates(f'section ({section!r}) must be in range of -4 to 19')

        # For better understanding of this code, read get_block()'s source

        if section is None or isinstance(section, int):
            section = self.get_section(section or 0)

        if section is None or 'block_states' not in section:
            air = Block.from_name('minecraft:air')
            for i in range(4096):
                yield air
            return

        states = section['block_states']['data']
        palette = section['block_states']['palette']

        bits = max((len(palette) - 1).bit_length(), 4)

        state = index // (64 // bits)

        data = states[state]
        if data < 0:
            data += 2**64

        bits_mask = 2**bits - 1

        offset = index % (64 // bits) * bits

        data_len = 64 - offset
        data >>= offset

        while index < 4096:
            if data_len < bits:
                state += 1
                new_data = states[state]
                if new_data < 0:
                    new_data += 2**64

                data = new_data
                data_len = 64

            palette_id = data & bits_mask
            yield Block.from_palette(palette[palette_id])

            index += 1
            data >>= bits
            data_len -= bits

    def stream_chunk(self) -> Generator[Block, None, None]:
        """
        Returns a generator for all the blocks in the chunk

        This is a helper function that runs Chunk.stream_blocks from section 0 to 15

        Yields
        ------
        :class:`anvil.Block`
        """
        for i in range(-4,20):
            for block in self.stream_blocks(section=i):
                yield block

    def get_tile_entity(self, x: int, y: int, z: int) -> Optional[nbt.TAG_Compound]:
        """
        Returns the tile entity at given coordinates, or ``None`` if there isn't a tile entity

        To iterate through all tile entities in the chunk, use :class:`Chunk.tile_entities`
        """
        for tile_entity in self.tile_entities:
            t_x, t_y, t_z = [tile_entity[k].value for k in 'xyz']
            if x == t_x and y == t_y and z == t_z:
                return tile_entity

    def get_biome(self, x: int, y: int, z: int) -> Biome:
        """
        Returns the biome given x, y, z
        """
        if x < 0 or x > 15:
            raise OutOfBoundsCoordinates(f'X ({x!r}) must be in range of 0 to 15')
        if z < 0 or z > 15:
            raise OutOfBoundsCoordinates(f'Z ({z!r}) must be in range of 0 to 15')
        if y < -64 or y > 319:
            raise OutOfBoundsCoordinates(f'Y ({y!r}) must be in range of -64 to 319')

        section = self.get_section(y // 16)
        if 'biomes' not in section:
            raise DataNotAvailable('Biomes are not available')
        
        biomes = section['biomes']
        biomes_palette = biomes['palette']

        # If there is only one biome, 'data' is not required
        if 'data' not in biomes:
            return Biome.from_name(biomes_palette[0].value)

        states = biomes['data']

        index = ((y % 16 // 4) * 4 * 4) + (z // 4) * 4 + (x // 4)
        bits = (len(biomes_palette) - 1).bit_length()
        state = index // (64 // bits)
        data = states[state]

        # makes sure the number is unsigned
        # by adding 2^64
        # could also use ctypes.c_ulonglong(n).value but that'd require an extra import
        if data < 0:
            data += 2**64

        shifted_data = data >> (index % (64 // bits) * bits)
        # TODO: fix here if index can span across 2 64bit numbers

        # get `bits` least significant bits
        # which are the palette index
        palette_id = shifted_data & 2**bits - 1
        return Biome.from_name(biomes_palette[palette_id].value)
